module S4U
	module S4u_offset
		class Offset
			def initialize()
				ext=IS_OSX ? ".pdf" : ".svg"
				cursor_path= File.join(File.dirname(__FILE__),("icon_mouse" + ext))
				@cursor_id = UI.create_cursor(cursor_path,0,0) if cursor_path
				@strings=S4U::S4u_offset::Strings
				@scale_factor= Sketchup.version.to_i > 17 ? UI.scale_factor : 1
				@snap=20*@scale_factor
				@list_type=[@strings["Both Side"],@strings["Reverse Side"],@strings["Random Offset"],@strings["Group Offset"]]
				@random_number=4

				@mode=0
				@double_side=0
				@random=0
				@reverse=0
				@setting= S4U::S4u_offset.instance_variable_get :@setting
				@mode,@double_side,@random,@reverse=@setting
				@random=0
				@reverse=0
				@last_tool=S4U::S4u_offset.instance_variable_get(:@last_tool)

			end
			def activate
				@edit_transform= axes_edit_transform
				@ip1 = Sketchup::InputPoint.new
				@ip2 = Sketchup::InputPoint.new
				@ip = Sketchup::InputPoint.new
				@times=0
				model=Sketchup.active_model
				model.active_view.lock_inference
				@sels=model.selection
				entities=model.active_entities
				@faces=@sels.grep(Sketchup::Face)
				@edge_faces=@faces.collect{|f| f.edges}.flatten.uniq
				edges=@sels.select{|e| e && e.valid? && e.is_a?(Sketchup::Edge) }.uniq
				@edges=edges-@edge_faces#-edge_curves
				if @edges.empty? && @faces.empty? #&& @curves.empty?
					UI.messagebox @strings["Select Edges or Faces"]
					model.select_tool nil
				end
				@faces_vertices=@faces.collect{|e| e.vertices}.flatten.uniq
				@edges_vertices=@edges.collect{|e| e.vertices}.flatten.uniq
				check_hide_edge=!Sketchup.active_model.rendering_options['DrawHidden']
				@list_edges= S4U::S4u_offset.find_weld(@edges,check_hide_edge)
				@list_faces={}
				@faces.each{|f| @list_faces[f] = S4U::S4u_offset.find_weld(f.edges,check_hide_edge) }
				@hash_face_of_edges={}
				fs = @edges.flat_map{|p| p.faces }
				if fs[0]
					fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
					fs=fs.sort_by{|f| f[1]}.reverse
					fs.each{|f| @hash_face_of_edges[f[0]]=f[1] }
				end
				update_selection

				self.reset
			end
			def update_selection()
				hidden=Sketchup.active_model.rendering_options['DrawHidden']
				@sels.clear
				if hidden
					@sels.add (@edges+@edge_faces+@faces).select{|e| e && e.valid?}
				else
					sels=(@edges+@edge_faces+@faces).select{|e| e && e.valid?}
					sels=sels - sels.select{|e| e.class==Sketchup::Edge && (e.soft? || e.hidden?)}
					@sels.add sels

				end
			end
			def axes_edit_transform
				return Sketchup.active_model.axes.transformation
			end
			def onSetCursor
				UI.set_cursor(@cursor_id) if @cursor_id
			end
			def getExtents
				bb = Sketchup.active_model.bounds
				bb.add @ip1.position if @ip1.valid?
				bb.add @ip2.position if @ip2.valid?
				return bb
			end
			def getMenu(menu, flags, x, y, view)
				menus=[]
				menus<< menu.add_item(@list_type[0]){
					@double_side=1-@double_side;
					make_offset
					update_selection
					set_status_text(view)
				}
				menu.set_validation_proc(menus[-1]) { (@double_side==0 )? MF_UNCHECKED : MF_CHECKED }
				menus<< menu.add_item(@list_type[1]){
					@reverse= 1-@reverse
					make_offset
					update_selection
					set_status_text(view) }
				menu.set_validation_proc(menus[-1]) { @reverse==0 ? MF_UNCHECKED : MF_CHECKED }
				menu.add_separator
				menus<< menu.add_item(@list_type[2]){
					@random= 1-@random
					make_offset
					update_selection
					set_status_text(view) }

				menu.set_validation_proc(menus[-1]) { (@random==0) ? MF_UNCHECKED : MF_CHECKED }
				menus<< menu.add_item(@list_type[3]){ @mode=1-@mode;set_status_text(view) }
				menu.set_validation_proc(menus[-1]) { @mode==0 ? MF_UNCHECKED : MF_CHECKED }

			end
			def deactivate(view)
				self.create_geometry
				S4U::S4u_offset.instance_variable_set(:@setting,[@mode,@double_side,@random,@reverse])
				S4U::S4u_offset.instance_variable_set(:@last_tool,@last_tool)
			end
			def get_vertices(face)
				return nil if not( face && (face.is_a?Sketchup::Face))
				ver=[]
				vertices=[]
				vertices =face.outer_loop.vertices*3
				(0..face.outer_loop.vertices.length-1).each { |i|
					j=face.outer_loop.vertices.length+i
					p1=vertices[j-1].position.clone
					p2=vertices[j].position.clone
					p3=vertices[j+1].position.clone
					v1=p1.vector_to p2
					v2=p2.vector_to p3
					ver.push vertices[j] if not( v1.samedirection? v2)
				}
				return ver
			end
			def get_vertices_from_vertices(vers_start)
				ver=[]
				if vers_start[0]==vers_start[-1]
					vertices =vers_start[0..-2]*3
					(0..vers_start.length-1).each { |i|
						j=vers_start.length+i
						p1=vertices[j-1].position.clone
						p2=vertices[j].position.clone
						p3=vertices[j+1].position.clone
						v1=p1.vector_to p2
						v2=p2.vector_to p3
						ver.push vertices[j] if  not( v1.samedirection? v2)
					}
				else
					vertices =vers_start*3
					ver<< vers_start[0]
					(1..vers_start.length-2).each { |i|
						j=vers_start.length+i
						p1=vertices[j-1].position.clone
						p2=vertices[j].position.clone
						p3=vertices[j+1].position.clone
						v1=p1.vector_to p2
						v2=p2.vector_to p3
						ver.push vertices[j] if  not( v1.samedirection? v2)
					}
					ver<< vers_start[-1]
				end
				return ver
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@snap) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip="       " + @strings.GetString("1/3")
					return ip
				end
				return nil
			end
			def center_face(face,num=200)
				return nil if face.outer_loop.vertices.length> num #|| face.loops.length>1
				vertices=get_vertices(face)
				return nil if !vertices
				numver=vertices.length
				cp=Geom::Point3d.new
				vertices.each{|v|
					cp.x=cp.x+v.position.x
					cp.y=cp.y+v.position.y
					cp.z=cp.z+v.position.z
				}
				cp.x=cp.x/ numver
				cp.y=cp.y/ numver
				cp.z=cp.z/ numver
				return cp
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				return nil if edge.length < (view.pixels_to_model(5*@snap, p1))
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@snap) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip= @strings.GetString("1/3")
					return ip
				end
				return nil
			end
			def get_bounds(element)
				b=nil
				if element.is_a?Sketchup::ComponentInstance
					b=element.definition.bounds
				elsif element.is_a?Sketchup::Group
					b=element.local_bounds
				elsif element.is_a?Sketchup::Image
					b=element.bounds
				else
					return nil
				end
				return (0..7).collect{|i| b.corner(i).clone.transform element.transformation } if b
			end
			def set_special_point(x,y,view,ip)
				ph = view.pick_helper
				ph.init(x, y)
				ph.do_pick(x, y)
				sels=view.model.selection.select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) || e.is_a?(Sketchup::Image))}
				points=[]
				@picked_element << ph.best_picked  if (e=ph.best_picked) && (e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)))
				@picked_element.select!{|e| e && e.valid?}
				sels+=@picked_element
				best_entity = sels.sort_by{|e| view.screen_coords(e.bounds.center).distance(view.screen_coords(ip.position)).to_f }
				best_entity=best_entity[0..7].select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) )}
				if !best_entity.empty?
					best_entity.each{|e|
						ps = get_bounds(e)
						points+=ps
						ps.permutation(2).to_a.each{|a|  points << Geom.linear_combination(0.5, a[0],0.5,a[1]) }
					}
				end
				points.uniq!
				picked_bounds = points.find_all{|p| ph.test_point(p,x,y,1.5*@snap) }
				if picked_bounds
					picked_bounds.each{|picked|
						model=view.model
						ray=view.pickray x, y
						if ray
							ray=[picked,ray[1].reverse!]
							item=model.raytest(ray,false)
							if !item
								@edge=true
								ip1=Sketchup::InputPoint.new(picked)
								ip.copy! ip1
								view.tooltip=@strings.GetString("BoundingBox")
								return picked
							end
						end
					}
				end
				if ip && ip.valid? && ip.face
					cf=center_face(ip.face)
					return ip if !cf
					cf.transform! ip.transformation
					ph = view.pick_helper
					ph.init(x, y)
					picked = ph.test_point(cf,x,y,@snap)
					if picked
						ip1=Sketchup::InputPoint.new(cf)
						@face=true
						view.tooltip=@strings.GetString("Center face")
						ip.copy! ip1
					else
						@edge=ip.face.edges.find{|e| set_pointonedge(x,y,view,ip,e) }
					end
				elsif ip.edge
					@edge=set_pointonedge(x,y,view,ip,ip.edge)
				end
				return ip
			end
			def onMouseMove(flags, x, y, view)
				@direction=nil
				@value_input=nil
				@shift_press= (CONSTRAIN_MODIFIER_MASK&flags==CONSTRAIN_MODIFIER_MASK)
				@ctrl_press= (COPY_MODIFIER_MASK&flags==COPY_MODIFIER_MASK)
				@alt_press= (ALT_MODIFIER_MASK&flags==ALT_MODIFIER_MASK)
				@face=@edge=nil
				@center_point=nil
				if(@state == 0 )
					@ip.pick view, x, y
					if @ip != @ip1
						view.tooltip = @ip.tooltip
						set_special_point(x,y,view,@ip) #if !view.inference_locked?
						view.invalidate
						@ip1.copy! @ip
						@ip2.clear
					end
					set_status_text(view)
				elsif (@state > 0 ) # !@nodraw
					@ip.pick view, x, y, @ip1
					view.tooltip = @ip.tooltip if( @ip.valid? )
					set_special_point(x,y,view,@ip) #if !view.inference_locked?
					if @shift_press && @p_shift && @ip!=@ip1 && @ip2!=@ip1
						p2_line=@ip.position.project_to_line([@ip1.position,@p_shift])
						@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
					elsif @ip_lock && @ip!=@ip1 && @ip2!=@ip1
						p2_line=@ip.position.project_to_line([@ip1.position,@ip_lock])
						@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
					elsif @ip.position.distance(@ip1.position)>0
						@ip2.copy! @ip
					end
					if @ip1 && @ip2 && @ip1.valid? && @ip2.valid?
						length = @ip1.position.distance(@ip2.position)
						@points=[@ip1.position,@ip2.position]
						set_status_text(view,length)
					end
				end
				view.invalidate
			end
			def onLButtonDoubleClick(flags, x, y, view)
				if @last_tool
					self.make_offset(@last_tool)
					update_selection
					set_status_text(view)
					view.model.select_tool nil
				end
			end
			def onLButtonDown(flags, x, y, view)
				if @state == 0
					if @ip1.valid?
						@state = 1
					end
				elsif @state == 1
					if @ip2.valid?
						if @alt_press
							create_geometry
						end
						self.reset(view)
					end
				end
				set_status_text(view)
			end
			def onKeyDown(key, repeat, flags, view)
				if( key == CONSTRAIN_MODIFIER_KEY )
					@shift_press =true
					if( view.inference_locked? )
						view.lock_inference
					elsif( @state == 0 && @ip1.valid? )
						view.lock_inference @ip1
					elsif( @state == 1 && @ip2.valid? )
						@p_shift=@ip.position
						view.lock_inference @ip2, @ip1
					end
				elsif key==VK_UP
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p=  @ip1.position
						p1=p + @edit_transform.zaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==VK_LEFT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p1=@ip1.position + @edit_transform.yaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==VK_RIGHT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p1=@ip1.position + @edit_transform.xaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==ALT_MODIFIER_KEY
					@alt_press=true
				elsif key==9
					@reverse=1-@reverse
					make_offset
					update_selection
					set_status_text(view)
				end
			end
			def onKeyUp(key, repeat, flags, view)
				if( key == CONSTRAIN_MODIFIER_KEY )
					@shift_press=nil
					@p_shift=nil
					view.lock_inference
				elsif key==ALT_MODIFIER_KEY
					@alt_press=nil
				end
			end
			def onUserText(text, view)

				if text.include?('*') && @times.abs>=1
					@times=(text.split('*')[1]).to_i
					self.make_offset
					update_selection
					set_status_text(view)
					return
				end
				if text.include?('/') && @times.abs>=1
					@times=-(text.split('/')[1]).to_i
					self.make_offset
					update_selection
					set_status_text(view)
					return
				end
				if @state==0
					value=cal(text)
					return if !value
					vec=X_AXIS.clone
					vec.length=value
					pt2 = @ip1.position + vec
					@ip2.copy! Sketchup::InputPoint.new(pt2)
					@value_input= value

					self.make_offset
					update_selection

					set_status_text(view)
				elsif @state == 1
					vec = @ip2.position - @ip1.position
					return if vec.length==0
					value=cal(text)
					return if !value
					vec.length = value
					pt2 = @ip1.position + vec
					set_status_text(view,value)
					@ip2.copy! Sketchup::InputPoint.new(pt2)
					@value_input= value
					self.make_offset
					update_selection

					set_status_text(view)
				end
			end
			def draw(view)
				if @ip
					if @ip.valid?
						@ip.draw(view)
					end
					if @face || @edge
						view.draw_points @ip.position, 9*@scale_factor, 1, "red"
					elsif @center_point
						view.draw_points @ip.position, 9*@scale_factor, 1, "red"
					end
				end
				if @ip1.valid? && @ip1.display?
					@ip1.draw(view)
				end
				if @state>0 && @ip2.valid?
					if @p_shift ||((@face || @edge) && @ip_lock)
						p=@p_shift
						p = @ip_lock if (@face || @edge) && @ip_lock
						p2_line=@ip.position.project_to_line([@ip1.position,p])
						view.draw_points p2_line, 9*@scale_factor, 2, "red"  if p2_line
						@ip2.copy! Sketchup::InputPoint.new(p2_line)
						view.line_stipple = "."
						view.drawing_color ="blue"
						view.draw_line(@ip.position, @ip2.position)
						view.line_stipple = ""
						view.drawing_color ="black"
					end
					@ip2.draw(view) if @ip2.display?
					view.set_color_from_line(@ip1, @ip2)
					view.line_width=2
					view.line_stipple = ""
					view.draw_line(@ip1.position, @ip2.position)
				end
				view.drawing_color ="black"
				view.line_width=1
				view.line_stipple = ""
				ps=[]
				if @state>0
					ps=make_offset()
				elsif @list_points
					ps=@list_points
				end
				ps.each{|pl|
					pl.each{|p|
						if p[1].length>0
							p[1].each{|li|
								view.draw_polyline(li[0..1]) if li.length>3
							}
						end
					} if ps.length>0
				}
			end
			def onReturn(view)
				view.model.select_tool nil
			end
			def onCancel(flag, view)
				if flag==2
					self.reset(view)
					view.model.select_tool nil
				end
				@list_points=nil
				self.reset(view)
			end
			def reset(view=nil)
				@state = 0
				@picked_element=[]
				@ip1.clear
				@ip2.clear
				@center=[]
				@ip_lock=nil
				@direction=nil
				@value_input=nil
				@selected_edge=nil
				set_status_text(view)
			end
			def set_status_text(view=nil,value=nil)
				text=@list_type[@mode] + " | "
				Sketchup::set_status_text @strings["Length"], SB_VCB_LABEL
				if @state == 0
					if @mode==0
						text += @strings["Select First Point or Input Length"]
					else
						text += @strings["Select First Point or Input Length"]
					end
				elsif @state == 1
					text += @strings["Select Second Point or Input Length"]
					text += " | " +  @strings["Multiple Offset : *number or /number"]
				end
				text += " | " +  @strings["Press TAB to Reverse Side"]
				if IS_OSX
					text += " | " +  @strings["Hold Command to Create Offset"]
				else
					text += " | " +  @strings["Hold Alt for to Create Offset"]
				end
				if value
					Sketchup::set_status_text value.to_s, SB_VCB_VALUE
				end
				Sketchup.status_text = text
				if view
					view.invalidate
					#view.lock_inference
				end
			end
			def cal(input)
				if op=input[/[\+\-\/\*]/]
					begin
					if input[/[\+\-]/]
						index= input.rindex(/[\+\-]/)
					elsif input[/[\*\/]/]
						index= input.rindex(/[\*\/]/)
					end
					if index
						if index>0
							s=[input[0..(index-1)],input[(index+1)..-1]]
						else
							s=["0",input[(index+1)..-1]]
						end
						op=input[index]
						if op=="+"
							value= cal(s[0]).to_s.to_f + cal(s[1]).to_s.to_f
						elsif op=="-"
							value= cal(s[0]).to_s.to_f - cal(s[1]).to_s.to_f
						elsif op=="*"
							value= cal(s[0]).to_s.to_f * cal(s[1]).to_s.to_f
						elsif op=="/"
							value= cal(s[0]).to_s.to_f / cal(s[1]).to_s.to_f
						end
					end
					rescue
						value=parse_length(input)
					end
				else
					value = parse_length(input)
				end
				return parse_length(value)
			end
			def parse_length(input)
				(return 0) if !input
				input=input.to_s
				if !Sketchup.parse_length("1.1")
					input.gsub!(/\./, ',')
				else
					input.gsub!(/\,/, '.')
				end
				begin
					value=input.to_l
				rescue
					value=input.to_f
				end
				return value
			end
			def get_point_from_face(faces,ps)
				result=[]
				vec=ps[1]-ps[0]
				faces.each{|face|
					face.loops.each{|loop|
						points=loop.vertices.collect{|p| p.position}
						points<<points[0]
						points1=points.collect{|p| p+vec}
						if (face.normal.angle_between(Z_AXIS)> Math::PI*0.5) || (face.normal.angle_between(vec)> Math::PI*0.5)
							points.reverse!
							points1.reverse!
						end
						result<<[points,face,(loop == face.outer_loop)]
					}
				}
				return result
			end
			def get_point_from_curve(cuvers,ps)
				result=[]
				vec=ps[1]-ps[0]
				cuvers.each{|curve|
					points=curve.vertices.collect{|p| p.position}
					points1=points.collect{|p| p+vec}
					result<<[points,points1]
				}
				return result
			end
			def get_point_from_edge(edges,ps)
				vec=ps[1]-ps[0]
				result=[]
				edges.each{|edge|
					next if ps[0].on_line?(edge.line) && ps[1].on_line?(edge.line)
					points=edge.vertices.collect{|p| p.position}
					points1=points.collect{|p| p+vec}
					result<<[points,points1]
				}
				return result
			end
			def get_direction(points=[@ip1.position, @ip2.position])
				view=Sketchup.active_model.active_view
				direction=nil
				if @faces.length>0
					face=@faces.find{|f|
						(f.classify_point(points[0])==Sketchup::Face::PointOnEdge ||
						f.classify_point(points[0])==Sketchup::Face::PointOnVertex)
					}
					if face
						direction=@faces.find{|f|
							(f.classify_point(points[0])==Sketchup::Face::PointOnEdge ||
							f.classify_point(points[0])==Sketchup::Face::PointOnVertex) &&
							(S4U::S4u_offset.point_in_polygon_2D(view.screen_coords(points[1]),(f.outer_loop.vertices.collect{|p| view.screen_coords(p.position)}),true))
						}
						return (direction ? -1 : 1)
					end
				end

				if @edges.length>0
					#edges=@edges
					#list= S4U::S4u_offset.find_weld(edges,false)
					direction=1
					list= @list_edges
					offset=[]
					if list[0]
						list.each{|list1|
							list1.each{|li|
								(0..li.length-2).each{|i|
									line=[li[i].position,li[i+1].position]
									if S4U::S4u_offset.point_in_polygon_2D(points[0],[li[i].position,li[i+1].position],true)
										@selected_edge=[li[i],li[i+1],li]
										if (li[0].position.distance(li[-1].position)).to_f <= 0.00001
											ps=li.collect{|p| (p.position)}
											direction=S4U::S4u_offset.point_in_polygon_2D(points[1],ps,true)
											return (direction ? -1 : 1)
										else
											direction=1
											fs=li[0].faces
											li.each{|p| fs= fs & p.faces}
											if fs.empty?
												# fs= li.flat_map{|p| p.faces }
												# if fs[0]
													# fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
													# fs=fs.sort_by{|f| f[1]}
													# fs=[fs[-1][0]]
												# end
											end
											if fs && fs.length>0
												# if fs.find{|f| (f.classify_point(points[0])==Sketchup::Face::PointOnEdge ||
														# f.classify_point(points[0])==Sketchup::Face::PointOnVertex) &&
														# (S4U::S4u_offset.point_in_polygon_2D(points[1],
														# (f.outer_loop.vertices.collect{|p| p.position}),true))
													# }
													# direction= -1
												# end
												if fs.find{|f| (f.classify_point(points[0])==Sketchup::Face::PointOnEdge ||
														f.classify_point(points[0])==Sketchup::Face::PointOnVertex) &&
														(S4U::S4u_offset.point_in_polygon_2D( view.screen_coords(points[1]),
														(f.outer_loop.vertices.collect{|p| view.screen_coords(p.position)}),true))
													}
													direction= -1
												end
											else
												if li.length==2
													vp=points[1]-points[0]
													v=line[1]-line[0]
													vec=v.cross vp
													vecz=S4U::S4u_offset.normal_vertices(line)
													direction= (vec.length!=0 && vec.samedirection?(vecz)) ? 1 : -1
												else
													line1=[view.screen_coords(li[i].position),view.screen_coords(li[i+1].position)]
													line1[0].z=0
													line1[1].z=0
													v=line1[1]-line1[0]
													p1=view.screen_coords(points[1])
													p1.z=0
													p=p1.project_to_line line1
													v1=p1-p
													vec=v1.cross v
													ps =li.collect{|p| view.screen_coords(p.position)}
													ps=ps.collect{|p| [p.x,p.y,0]}
													vecz=S4U::S4u_offset.normal_vertices(ps)
													direction= (vec.length!=0 && vec.samedirection?(vecz)) ? 1 : -1
												end
											end
											return direction
										end
									end
								}
							}
						}
					end
				end
				return 1
			end
			def make_offset(last_tool=nil)
				ps=[]
				if last_tool
					vec,@direction,@times,points=last_tool
				else
					points=@points
					if !@points || !@points[0] || !@points[1]
						if @ip1 && @ip2 && @ip1.valid? && @ip2.valid? && @ip1.position.distance(@ip2.position)>0
							@points=[ @ip1.position, @ip2.position]
							 points=@points
						else
							return []
						end
					end

					vec=points[1]-points[0]
					return [] if vec.length==0

					if vec.length!=0 && @value_input && @value_input!=0
						vec.length=@value_input
						points[1]=points[0]+vec
					end
					direct = get_direction(points)
					@direction=direct if direct
					if @value_input
						@direction=(@value_input>0) ? @direction:-@direction
					end
					@last_tool=[vec,@direction,@times,points]
				end
				distance=points[0].distance(points[1])
				dis=distance
				p1=points[0]
				return [] if vec.length==0
				@times=1 if @times==0
				(dis=(distance/@times.to_f).abs) if @times<0
				times=@times.abs
				(1..times).each{|t|
					if (@random==1) && t>1
						if(t==times) && (@times<0)
							vec.length=distance
						else
							index_random=rand(@random_number)
							vec.length=vec.length+dis*(index_random)
							next if (vec.length > distance) &&  @times<0
						end
					else
						vec.length=t*dis
					end
					p2=p1+vec
					ps+=@faces.collect{|face| get_offset(face,[p1,p2],@direction) }
					ps<<get_offset_edge(@edges,[p1,p2],@direction)
				}
				# ps.each_index{|i| ps[i].each_index{|i1| ps[i][i1][1].select!{|p|
					# (p[1]-p[0]).length!=0 && (p[3].position-p[2].position).length!=0 &&
					# (p[1]-p[0]).samedirection?(p[3].position-p[2].position)}}}
				@list_points=ps.clone
				return ps
			end
			def get_offset_edge(edges,points,direction)
				vec=points[1]-points[0]
				return [] if vec.length==0
				check_hide_edge=!Sketchup.active_model.rendering_options['DrawHidden']
				list=@list_edges
				# list =S4U::S4u_offset.find_weld(edges,check_hide_edge)
				offset=[]
				if list[0]
					list.each{|list1|
						list1.each{|li|
							border1=border2=nil
							if check_hide_edge
								ed= li[0].edges & edges
								if ed.length==2 && (edge=ed.find{|e| e.soft? || e.hidden?  } )
									border1=[edge.start.position,edge.end.position]
								end
								ed= li[-1].edges & edges
								if ed.length==2 && (edge=ed.find{|e| e.soft? || e.hidden?  } )
									border2=[edge.start.position,edge.end.position]
								end
							end
							ps=li.collect{|p| p.position}
							vecz=S4U::S4u_offset.normal_vertices(ps)

							if @selected_edge && @selected_edge[2]==li
								if ps.length==2
									v=ps[1]-ps[0]
									vect=v.cross vec
									vecz=vect if vect.length!=0
								else
										fs=li[0].faces
										li.each{|p| fs= fs & p.faces}
										if fs.empty?
											# fs= li.flat_map{|p| p.faces }.uniq
											# if fs[0]
												# fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
												# fs=fs.sort_by{|f| f[1]}
												# fs=[fs[-1][0]]
											# end
										end
										if fs && fs.length>0
											fs.select!{|f| @hash_face_of_edges[f] }
											face=fs.max_by{|f| @hash_face_of_edges[f] }
											vecz=face.normal.clone
											vecz.reverse! if S4U::S4u_offset.check_direction_face(face,ps[0..1])[0]!=direction
										else
											if S4U::S4u_offset.check_direction_edge(ps,ps[0..1])[0]!=direction
												vecz.reverse!
											end
										end
									# vecz.reverse! if S4U::S4u_offset.check_direction_edge(ps,ps[0..1])[0]!=direction
								end

							else
								# if ps.length==2
									# fs=li[0].faces
									# li.each{|p| fs= fs & p.faces}
									# if fs.empty?
										# fs= li.flat_map{|p| p.faces }
										# if fs[0]
											# fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
											# fs=fs.sort_by{|f| f[1]}
											# fs=fs[-1][0]
										# end
									# end
									# if fs[0]
										# face=fs[0]
										# vecz=face.normal.clone
										# vecz.reverse! if S4U::S4u_offset.check_direction_face(face,ps[0..1])[0]!=direction
									# else
										# v=ps[1]-ps[0]
										# vect=v.cross vec
										# vecz=vect if vect.length!=0
									# end
								# else
									close=( ps[0].distance(ps[-1]).to_f <=0.00001)
									if close
										vecz.reverse! if S4U::S4u_offset.check_direction_edge(ps,ps[0..1])[0]!=direction
									else
										fs=li[0].faces
										li.each{|p| fs= fs & p.faces}
										if fs.empty?
											# fs= li.flat_map{|p| p.faces }.uniq
											# if fs[0]
												# fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
												# fs=fs.sort_by{|f| f[1]}
												# fs=[fs[-1][0]]
											# end
										end
										if fs && fs.length>0
											fs.select!{|f| @hash_face_of_edges[f] }
											face=fs.max_by{|f| @hash_face_of_edges[f] }
											vecz=face.normal.clone
											vecz.reverse! if S4U::S4u_offset.check_direction_face(face,ps[0..1])[0]!=direction
										else

											if S4U::S4u_offset.check_direction_edge(ps,ps[0..1])[0]!=direction
												vecz.reverse!
											end
										end
									end
								# end
							end
							vecz.reverse! if @reverse==1
							pt=S4U::S4u_offset.offset_points_border(li,vec,vecz,border1,border2)
							offset << [li,pt]
							if @double_side==1
								pt=S4U::S4u_offset.offset_points_border(li,vec,vecz.reverse,border1,border2)
								offset << [li,pt]
							end
						}
					}
				end
				return offset
			end
			def get_offset(face,points,direction)
				check_hide_edge=!Sketchup.active_model.rendering_options['DrawHidden']
				list= @list_faces[face]
				#list=S4U::S4u_offset.find_weld(face.edges,check_hide_edge)
				offset=[]
				if list[0]
					list[0].each{|li|
						ed= li[0].edges & face.edges
						border1=border2=nil
						if check_hide_edge
							if ed.length==2 && (edge=ed.find{|e| e.soft? || e.hidden?  } )
								border1=[edge.start.position,edge.end.position]
							end
							ed= li[-1].edges & face.edges
							if ed.length==2 && (edge=ed.find{|e| e.soft? || e.hidden?  } )
								border2=[edge.start.position,edge.end.position]
							end
						end
						#li=get_vertices_from_vertices(li)
						ps=li.collect{|p| p.position}
						vec=points[1]-points[0]
						vecz=face.normal.clone
						vecz.reverse! if S4U::S4u_offset.check_direction_face(face,ps[0..1])[0]!=direction
						if vec.length>0
							vecz.reverse! if @reverse==1
							pt=S4U::S4u_offset.offset_points_border(li,vec,vecz,border1,border2)
							offset << [li,pt]
							if @double_side==1
								pt=S4U::S4u_offset.offset_points_border(li,vec,vecz.reverse,border1,border2)
								offset << [li,pt]
							end
						end
					}
				end
				return offset
			end
			def create_geometry()
				return if !@list_points || @list_points.empty?
				ps = @list_points
				model=Sketchup.active_model
				entities=model.active_entities
				sels=model.selection
				model.start_operation(TOOL,true,false,false)
				edges_face=@faces.grep(Sketchup::Face).collect{|f| f.loops.collect{|l| l.edges}}.flatten
				@group=Sketchup.active_model.active_entities.add_group
				entities=@group.entities
				#ps=make_offset()
				vertices = (@faces_vertices + @edges_vertices)
				ps.each{|pl|
					pl.each{|p|
						g1=@group.entities.add_group
						curs=[]
						curve=nil
						curve_hash={}
						p[1].each{|li|
							if li.length>3
								ed=[]
								v1=li[2]
								v2=li[3]
								(ed=(v1.edges & v2.edges) )if v1 && v2
								if ed[0] && ed[0].curve
									curve=ed[0].curve
									curve_hash[curve]=[] if !curve_hash[curve]
									curve_hash[curve]<<li[0..1]
								else
									g1.entities.add_edges li[0..1]
								end
							end
						} if p[1].length>0
						if curve_hash.length>0
							curve_hash.keys.each{|key|
								cur= curve_hash[key]*2
								stop=true
								(0..curve_hash[key].length-1).each{|i|
									if !(cur[i][1]==cur[i+1][0])
										ps=cur.slice(i+1,curve_hash[key].length)
										#ps1=ps.flatten.uniq
										ps1=[ps.shift]
										count=0
										length=ps.length
										while(( ps.length>0) && (count<length))
											ps.each_index{|i|
												if ps1[-1][1].distance(ps[i][0]).to_f <= 0.0001
													ps1<<ps[i]
													ps.delete_at(i)
													break
												end
											}
											count+=1
										end
										g1.entities.add_curve ps1.flatten.uniq
										stop=false
										break
									end
								}
								if stop
									ps= curve_hash[key]
									ps1=[ps.shift]
									count=0
									length=ps.length
									while ((ps.length>0) && (count<length))
										ps.each_index{|i|
											if ps1[-1][1].distance(ps[i][0]).to_f <= 0.0001
												ps1<<ps[i]
												ps.delete_at(i)
												break
											end
										}
										count+=1
									end
									g1.entities.add_curve ps1.flatten.uniq
								end
							}
						end
					}
				}
				if @group && @group.valid? && @group.is_a?(Sketchup::Group)
					if @mode==0
						group=@group.explode
						group.each{|g| g.explode if g && g.valid? && g.is_a?(Sketchup::Group)}
					else
						@group.entities.each{|g| g.explode if g && g.valid? && g.is_a?(Sketchup::Group)}
					end
				end
				model.commit_operation
				sels.clear
				sels.add (@edges+@edge_faces+@faces).select{|e| e && e.valid?}
				rescue=>error
				p error
				p error.backtrace.join('\n')
				model.abort_operation
			end
		end
		def self.get_info_vertices(vers)
			close = (vers[0].position.distance(vers[-1].position).to_f <= 0.00001)
			angles={}
			if close
				vers1=[vers[-2]] + vers + [vers[1]]
				(1..vers.length-1).each{|i|
					v1=vers1[i-1].position-vers1[i].position
					v2=vers1[i+1].position-vers1[i].position
					v1.length=1
					v2.length=1
					v=v1+v2
					if v1.parallel?(v2)
						v=get_ocs(v).xaxis
						p=vers1[i].position
						plane=[vers1[i].position,(vers1[i+1].position-p)]
						plane1=[plane[0],plane[1].reverse]
						angles[vers1[i]]=[plane,plane1,v,(i-1)]
					else
						p=vers1[i+1].position.project_to_line [vers1[i].position,v]
						plane=[vers1[i].position,(vers1[i+1].position-p)]
						plane1=[plane[0],plane[1].reverse]
						angles[vers1[i]]=[plane,plane1,v,(i-1)]
					end
				}
			else
				v=vers[1].position-vers[0].position
				plane=[vers[0].position,v]
				angles[vers[0]]=[plane,v,0]
				vers1=vers
				(1..vers.length-2).each{|i|
					v1=vers1[i-1].position-vers1[i].position
					v2=vers1[i+1].position-vers1[i].position
					v1.length=1
					v2.length=1
					v=v1+v2
					if v1.parallel?(v2)
						v=get_ocs(v).xaxis
						p=vers1[i].position
						plane=[vers1[i].position,(vers1[i+1].position-p)]
						plane1=[plane[0],plane[1].reverse]
						angles[vers1[i]]=[plane,plane1,v,(i-1)]
					else
						p=vers1[i+1].position.project_to_line [vers1[i].position,v]
						plane=[vers1[i].position,(vers1[i+1].position-p)]
						plane1=[plane[0],plane[1].reverse]
						angles[vers1[i]]=[plane,plane1,v,i]
					end
				}
				v=vers[-1].position-vers[-2].position
				plane=[vers[-1].position,v]
				plane1=[vers[-1].position,v.reverse]
				angles[vers[-1]]=[plane,plane1,v,(vers.length-1)]
			end
			return angles
		end
		def self.get_ocs(z_axis)
			  if (z_axis.x.abs < (1.0/64.0)) && (z_axis.y.abs < (1.0/64.0))
				x_axis = Y_AXIS*z_axis
			  else
				x_axis = Z_AXIS*z_axis
			  end
			  y_axis = z_axis*x_axis
			  return Geom::Transformation.axes(ORIGIN, x_axis, y_axis, z_axis)
			  rescue => error
			  return Geom::Transformation.new(z_axis)
		end
		def self.normal_vertices(ver)
				return Z_AXIS if ver.length<2
				vx=nil;vy=nil;line=nil
				(0..ver.length-2).each{|i|
					vx=ver[i].vector_to ver[i+1]
					line=[ver[i],ver[i+1]]
					break if vx.length!=0
				}
				if line
					(0..ver.length-1).each{|i|
						v2=ver[i].project_to_line line
						vy=v2.vector_to ver[i]
						break if vy.length!=0
					}
					v= (vx.cross vy)
					return v if vx && vy && v.length!=0
				end
				trans=get_ocs(ver[1]-ver[0])
				return trans.yaxis
				if ver[0].z==ver[1].z
					return Z_AXIS.clone
				elsif ver[0].y==ver[1].y
					return Y_AXIS.clone
				else
					return X_AXIS.clone
				end
			end
		def self.check_direction_edge(ps,points)
			view=Sketchup.active_model.active_view
			if (points.length>1) && (points[1]!=points[0])
				v=points[1]-points[0]
				vz=normal_vertices ps
				vy=vz.cross v
				if vy.length!=0
					vy.length=points[0].distance(points[1])*0.1
					pm=Geom::Point3d.new (points[0].x+points[1].x)*0.5,(points[0].y+points[1].y)*0.5,(points[0].z+points[1].z)*0.5
					p = pm + vy
					if ps.length>1 && (ps[0].distance(ps[-1]).to_f <= 0.00001)
						if self.point_in_polygon_2D(p,ps,true)
							return [-1,vy]
						else

							return [1,vy.reverse]
						end
					end
					# direction=1
					# vec=nil
					# if ps.length==2
						# vp=points[1]-points[0]
						# v=ps[1]-ps[0]
						# vec=v.cross vp
						# vecz=S4U::S4u_offset.normal_vertices(ps)
						# direction= (vec.length!=0 && vec.samedirection?(vecz)) ? 1 : -1
					# else
						# line1=[view.screen_coords(points[0]),view.screen_coords(points[1])]
						# v=line1[1]-line1[0]
						# p1=view.screen_coords(points[1])
						# p=p1.project_to_line line1
						# v1=p1-p
						# vec=v1.cross v
						# vecz=S4U::S4u_offset.normal_vertices(ps)
						# direction= (vec.length!=0 && vec.samedirection?(vecz)) ? 1 : -1
					# end
					# return [direction,vec]
					p1=p.project_to_line ps[0..1]
					v1=p-p1
					if v1.length!=0 && v1.samedirection?(vy)
						return [1,v1]
					else
						return [-1,v1.reverse]
					end
				end
			end
			return [0,Z_AXIS.clone]
		end
		def self.point_in_polygon_2D(p,ps,check_border)
			if ps.length>2
				vz= normal_vertices ps
				vz=Z_AXIS if !vz || vz.length==0
				t=Geom::Transformation.new vz
				ps1=ps.collect{|pt| pt.transform t.inverse}
				p1=p.transform t.inverse
				return  Geom::point_in_polygon_2D(p1,ps1,check_border)
			elsif ps.length==2
				if p.on_line?(ps)
					if check_border
						return (p.distance(ps[0])<=ps[0].distance(ps[1]) && p.distance(ps[1])<=ps[0].distance(ps[1]) )
					else
						return (p.distance(ps[0])<ps[0].distance(ps[1]) && p.distance(ps[1])<ps[0].distance(ps[1]) )
					end
				else
					return false
				end
			else
				if check_border
					return (p.distance(ps[0])<=0.00001)
				else
					return (p.distance(ps[0])>0)
				end
			end
		end
		def self.check_direction_face(face,points1)
			view=Sketchup.active_model.active_view
			points=points1.collect{|p| p.project_to_plane(face.plane)}
			if (points.length>1) && (points[1]!=points[0]) && face
				v=points[1]-points[0]
				vy=face.normal.cross v
				if vy.length!=0
					vy.length=points[0].distance(points[1])*0.01
					pm=Geom::Point3d.new (points[0].x+points[1].x)*0.5,(points[0].y+points[1].y)*0.5,(points[0].z+points[1].z)*0.5
					p = pm + vy
					if face.classify_point(p)==1
						return [-1,vy]
					else
						return [1,vy.reverse]
					end
				end
			end
			return [0,face.normal]
		end
		def self.intersect_points_points(points,points1)
			(0..points.length-2).each{|i|
				line=[points[i+1],points[i]]
				(0..points1.length-2).each{|j|
					p = Geom::intersect_line_line(line,points1[j..j+1])
					if p && ((p.distance(line[0]) + p.distance(line[-1])) == line[0].distance(line[-1]))
						return [p,i]
					end
				}
			}
			return nil
		end
		def self.get_angle_from_points(vertices,vecz)
			ps=vertices.collect{|v| v.position}
			v1=ps[0]-ps[1]
			v2=ps[3]-ps[2]
			ang=v1.angle_between v2
			ang=ang%(Math::PI)
			vec=v2.cross v1
			if vec.length!=0
				ang= vec.samedirection?(vecz) ? ang : (2*Math::PI-ang)
			else
				ang= v1.samedirection?(v2) ? 0 : Math::PI
			end
			return ang
		end
		def self.point_on_plane(point,plane)
			p1=point.clone
			t_plane=Geom::Transformation.new(plane[0],plane[1])
			p1.transform! t_plane.inverse
			return 0 if p1.z.abs<0.00001
			return p1.z
		end
		def self.get_plane_from_intersect_vertices(vers1)
			line1=[vers1[0].position,vers1[1].position]
			line2=[vers1[2].position,vers1[3].position]
			v1=vers1[1].position-vers1[0].position
			v2=vers1[3].position-vers1[2].position
			v1.length=1
			v2.length=1
			v=v1+v2
			p =Geom::intersect_line_line line1,line2
			p=vers1[1].position if !p
			p1=vers1[0].position.project_to_line [p,v]
			plane=[p,(vers1[0].position-p1)]
			p2=vers1[3].position.project_to_line [p,v]
			plane1=[p,(vers1[3].position-p2)]
			return [plane,plane1,p]
		end
		def self.offset_points_border(vertices_start,vec,vecz,border1=nil,border2=nil,angle=Math::PI)
			return [] if vec.length==0
			vertices_info=get_info_vertices(vertices_start)
			angle= Math::PI
			vertices=vertices_start.clone
			points_start=vertices.collect{|p| p.position}
			points=points_start.clone
			points1=[]
			close=(points[0]==points[-1])

			if points.length==1
				points[0]=points[0]+vec.clone
			elsif points.length==2
				pt=[]
				v=points[1]-points[0]
				v=vecz.cross v
				if v.length!=0
					v.length=vec.length
					pt<<(points[0]+v.clone)
					pt<<(points[1]+v.clone)
					if (pt[1]-pt[0]).samedirection?(points[1]-points[0])
						points1<<[pt[0],pt[1],vertices[0],vertices[1]]
					end
				end
			elsif points.length>2
				pt=[]
				if !close
					i=1
					vp1=points[1]-points[0]
					v1=vecz.cross vp1
					if v1.length!=0
						v1.length=vec.length
						line1=[(points[i-1]+v1),((points[i]+v1)-(points[i-1]+v1))]
						v2=points[i+1]-points[i]
						v2=vecz.cross v2
						if v2.length!=0
							v2.length=vec.length
							vs=vertices[i-1..i] + vertices[i..i+1]
							ang=get_angle_from_points(vs,vecz)
							if ((ang-Math::PI).abs<0.00001)
								if ((points[i]+v2)-(points[i-1]+v1)).samedirection?(vertices[i].position-vertices[i-1].position)
									points1<<[(points[i-1]+v1),(points[i]+v2),vertices[i-1],vertices[i]]
								end
							else
								line2=[(points[i]+v2),((points[i+1]+v2)-(points[i]+v2))]
								p=Geom::intersect_line_line(line1,line2)
								if p && (p-(points[i-1]+v1)).length>0 &&
									((p-(points[i-1]+v1)).samedirection?(vertices[i].position-vertices[i-1].position))
									if ang>angle
										points1<<[(points[i-1]+v1),p,vertices[i-1],vertices[i]]
									else
										plane1=vertices_info[vertices[i-1]][0]
										plane2=vertices_info[vertices[i]][1]
										if point_on_plane(p,plane1)>=0 && point_on_plane(p,plane2)>=0
											points1<<[(points[i-1]+v1),p,vertices[i-1],vertices[i]]
										end
									end
								end
							end

						end
					end
				else
					# length=0
					# imax=0
					# (0..points.length-2).each{|i|
						# lt=points[i].distance points[i+1]
						# if length<lt
							# length=lt
							# imax=i
						# end
					# }
					# points= points[0..-2]*3
					# points=points.slice(imax,points_start.length-1)
					# vertices= vertices[0..-2]*3
					# vertices=vertices.slice(imax,points_start.length-1)
					# points=  [points[-1]] + points + [points[0]] + [points[1]]
					# vertices =  [vertices[-1]] + vertices + [vertices[0]]+ [vertices[1]]
					points=  [points[-2]] + points + [points[1]]
					vertices=  [vertices[-2]] + vertices + [vertices[1]]
				end
				index=1
				(index..points.length-3).each{|i|
					pt=[]
					v1=points[i]-points[i-1]
					v1=vecz.cross v1
					if v1.length!=0
						v1.length=vec.length
						line1=[(points[i-1]+v1),(points[i]+v1)]
						vs=vertices[i-1..i]  + vertices[i..i+1]
						lv1=[vertices[i-1].position,vertices[i].position]
					else
						next
					end
					ang=get_angle_from_points(vs,vecz)
					if (ang-Math::PI).abs<0.00001
						pt << (points[i]+v1)
					else
						vp2=points[i+1]-points[i]
						v2=vecz.cross vp2
						if v2.length!=0
							v2.length=vec.length
							line2=[(points[i]+v2),((points[i+1]+v2)-(points[i]+v2))]
							p=Geom::intersect_line_line(line1,line2)
							if p
								if ang>angle
									pt << p
								else
									plane1=vertices_info[vertices[i]][0]
									plane2=vertices_info[vertices[i+1]][1]
									if point_on_plane(p,plane1)>=0 && point_on_plane(p,plane2)>=0
										pt<<p
									end
								end
							end
						end
					end
					if pt.length==1
						vp1=points[i+1]-points[i]
						v1=vecz.cross vp1
						if v1.length!=0
							v1.length=vec.length
							line1=[(points[i]+v1),vp1]
							vs=vertices[i..i+1] + vertices[i+1..i+2]
							ang=get_angle_from_points(vs,vecz)
							if (ang-Math::PI).abs<0.00001
								pt<<(points[i+1]+v1)
							else
								vp2=points[i+2]-points[i+1]
								v2=vecz.cross vp2
								if v2.length!=0
									v2.length=vec.length
									line2=[(points[i+1]+v2),vp2]
									p=Geom::intersect_line_line(line1,line2)
									if p
										if ang>angle
											pt<<p
										else
											plane1=vertices_info[vertices[i]][0]
											plane2=vertices_info[vertices[i+1]][1]
											if point_on_plane(p,plane1)>=0 && point_on_plane(p,plane2)>=0
												pt<<p
											end
										end
									end
								end
							end
						end
						if pt.length==2
							stop=false
							length_count=points1.length
							if length_count>0 && (pt[0].distance(points1[-1][1]).to_f >= 0.0001)
								count=0
								points11=points1.collect{|p| p.clone }
								line2=[pt[0],pt[1]]
								while (points11.length>0) && (!stop) && (count<length_count)
									count+=1
									line1=[points11[-1][0],points11[-1][1]]
									p=Geom::intersect_line_line(line1,line2)
									plane1=vertices_info[points11[-1][2]][0]
									plane2=vertices_info[points11[-1][3]][1]
									#plane3=vertices_info[vertices[i]][0]
									#plane4=vertices_info[vertices[i+1]][1]
									# p1=pt[0].project_to_line [vertices[i].position,vertices[i+1].position]
									# plane5=[vertices[i].position,(pt[0]-p1)]
									if p
										if point_on_plane(p,plane1)>=0 && point_on_plane(p,plane2)>=0
											#point_on_plane(p,plane3)>=0 && point_on_plane(p,plane4)>=0
											pt[0]=p
											stop=true
										end
									end
									if !stop
										points11.delete_at(-1)
									end
								end
								if stop && (points11.length>0)
									if pt[0].distance(pt[1])>0 && (pt[1]-pt[0]).samedirection?(vertices[i+1].position-vertices[i].position)
										points11[-1][1]=pt[0]
										points1=points11
										points1<<[pt[0],pt[1],vertices[i],vertices[i+1]]
									end
								else
									stop=false
									count=0
									points11=points1.collect{|p| p.clone }
									while (points11.length>0) && (!stop) && (count==0)&&
										(pt[0].distance(points11[-1][1]).to_f >= 0.0001)
										count+=1
										line1=[points11[-1][0],points11[-1][1]]
										p=Geom::intersect_line_line(line1,line2)
										plane1=vertices_info[points11[-1][3]][0]
										plane2=vertices_info[vertices[i+1]][1]
										if p
											if  point_on_plane(p,plane1)>=0 &&  point_on_plane(p,plane2)>=0
												pt[0]=p
												stop=true
											end
										end
										if !stop
											points11.delete_at(-1)
										end
									end
									if stop && points11.length>0
										if  pt[0].distance(pt[1])>0 && (pt[1]-pt[0]).samedirection?(vertices[i+1].position-vertices[i].position)
											points11[-1][1]=pt[0]
											points1=points11
											points1<<[pt[0],pt[1],vertices[i],vertices[i+1]]
										end
									end
								end
							else
								if  pt[0].distance(pt[1])>0 && (pt[1]-pt[0]).samedirection?(vertices[i+1].position-vertices[i].position)
									points1<<[pt[0],pt[1],vertices[i],vertices[i+1]]
								end
							end

						end
					end
				}

				if !close
					stop=false
					length_count=points1.length
					count=0
					while (points1.length>0) && (!stop) && (count<length_count)
						count+=1
						vp1=points1[-1][1]-points1[-1][0]
						line1=[points1[-1][0],points1[-1][1]]
						vs=points1[-1][2..3] + vertices[-2..-1]
						vp2=points[-1]-points[-2]
						v2=vecz.cross vp2
						if v2.length!=0
							v2.length=vec.length
							ang=get_angle_from_points(vs,vecz)
							if ((ang-Math::PI).abs<0.00001)
								points1<<[(points[-2]+v2),(points[-1]+v2),vertices[-2],vertices[-1]]
								stop=true
							else
								line2=[(points[-2]+v2),((points[-1]+v2)-(points[-2]+v2))]
								p=Geom::intersect_line_line(line1,line2)
								plane1=vertices_info[points1[-1][2]][0]
								plane2=vertices_info[points1[-1][3]][1]
								if p && ((points[-1]+v2)-p).length>0 &&
									(((points[-1]+v2)-p).samedirection?(vertices[-1].position-vertices[-2].position))
									if point_on_plane(p,plane1)>=0 && point_on_plane(p,plane2)>=0
										points1[-1][1]=p if points1[-1]
										points1<<[p,(points[-1]+v2),vertices[-2],vertices[-1]]
										stop=true
									else
										plane1=vertices_info[vertices[-1]][1]
										plane2=vertices_info[points1[-1][2]][0]
										if point_on_plane(p,plane1)>=0 && point_on_plane(p,plane2)>=0
											points1[-1][1]=p if points1[-1]
											points1<<[p,(points[-1]+v2),vertices[-2],vertices[-1]]
											stop=true
										else
											points1.delete_at(-1)
										end
									end
								end
							end
						end
						break if stop
					end
				elsif points1.length>0
					points11=[]
					if (points1[0][0].distance(points1[-1][1]).to_f >= 0.00001)
						j=points1.length/2
						ps1=points1[0..j]
						ps2=points1[j+1..points1.length-1]
						if ps1.length>0 && ps2.length>0
							0.upto(ps1.length-1).each{|i1|
								line1=[ps1[i1][0],(ps1[i1][1]-ps1[i1][0])]
								line11=[ps1[i1][0],ps1[i1][1]]
								(ps2.length-1).downto(0).each{|i2|
									line2=[ps2[i2][0],(ps2[i2][1]-ps2[i2][0])]
									line22=[ps2[i2][0],ps2[i2][1]]
									if line1[1].length>0 && line2[1].length>0
										p= Geom::intersect_line_line line1,line2
										if p &&
											# (((p.distance(ps1[i1][0])<=ps1[i1][0].distance(ps1[i1][1])) &&
											  # (p.distance(ps1[i1][1])<=ps1[i1][0].distance(ps1[i1][1]))) ) &&
											# (((p.distance(ps2[i2][0])<=ps2[i2][0].distance(ps2[i2][1])) &&
											  # (p.distance(ps2[i2][1])<=ps2[i2][0].distance(ps2[i2][1]))) )
											p.on_line?(line11) && p.on_line?(line22) &&
												(S4U::S4u_offset.point_in_polygon_2D(p,line11,true) ||
												 S4U::S4u_offset.point_in_polygon_2D(line11[0],[line11[1],p],true)) &&
												(S4U::S4u_offset.point_in_polygon_2D(p,line22,true) )
											ps1[i1][0]=p
											ps2[i2][1]=p
											points11= ps1[i1..-1] + ps2[0..i2]
											stop=true
											break
										end

									end
								}
								if stop
									break
								end
							}
							if !stop
								0.upto(ps1.length-1).each{|i1|
									line1=[ps1[i1][0],(ps1[i1][1]-ps1[i1][0])]
									line11=[ps1[i1][0],ps1[i1][1]]
									(ps2.length-1).downto(0).each{|i2|
										line2=[ps2[i2][0],(ps2[i2][1]-ps2[i2][0])]
										line22=[ps2[i2][0],ps2[i2][1]]
										if line1[1].length>0 && line2[1].length>0
											p= Geom::intersect_line_line line1,line2
											if p &&
												p.on_line?(line11) && p.on_line?(line22) &&
													(S4U::S4u_offset.point_in_polygon_2D(p,line11,true)) &&
													(S4U::S4u_offset.point_in_polygon_2D(p,line22,true) ||
													S4U::S4u_offset.point_in_polygon_2D(line22[1],[line22[0],p],true))
												ps1[i1][0]=p
												ps2[i2][1]=p
												points11= ps1[i1..-1] + ps2[0..i2]
												stop=true
												break
											end

										end
									}
									if stop
										break
									else

									end
								}
							end
							if !stop
								i1=0
								i2=-1
								line1=[ps1[i1][0],(ps1[i1][1]-ps1[i1][0])]
								line2=[ps2[i2][0],(ps2[i2][1]-ps2[i2][0])]
								if line1[1].length>0 && line2[1].length>0
									p= Geom::intersect_line_line line1,line2
									if p
										(((p.distance(ps1[i1][0])<=ps1[i1][0].distance(ps1[i1][1])) &&
										  (p.distance(ps1[i1][1])<=ps1[i1][0].distance(ps1[i1][1]))) ||
										 ((p.distance(ps1[i1][0])<=p.distance(ps1[i1][1])) &&
										  (ps1[i1][0].distance(ps1[i1][1])<=p.distance(ps1[i1][1])))) &&
										(((p.distance(ps2[i2][0])<=ps2[i2][0].distance(ps2[i2][1])) &&
										  (p.distance(ps2[i2][1])<=ps2[i2][0].distance(ps2[i2][1]))) ||
											((p.distance(ps2[i2][1])<=p.distance(ps2[i2][0])) &&
											(ps2[i2][0].distance(ps2[i2][1])<=p.distance(ps2[i2][0])))
										)
										ps1[i1][0]=p
										ps2[i2][1]=p
										points11= ps1[i1..-1] + ps2[0..i2]
									end
								end
							end
						end
						points1 = points11 if points11.length>0
					end
				end
				return [] if (points1.length<=1)
			end
			if border1 && border1[0] && border1[1] && points1.length>0
				pl=[border1[0],(border1[0].vector_to border1[1])]
				p=Geom::intersect_line_line([points1[0][0],points1[0][1].vector_to(points1[0][0])],pl)
				points1[0][0]=p if p
			end
			if border2 && border2[0] && border2[1] && points1.length>0
				pl=[border2[0],(border2[0].vector_to border2[1])]
				p=Geom::intersect_line_line([points1[-1][0],(points1[-1][0].vector_to(points1[-1][1]))],pl)
				points1[-1][1]=p if p
			end
			points1=self.edit_points(points1,close)
			return points1
		end
		def self.edit_points(points1,close)
			return points1 if points1.length<1
			index=0
			points2=[]
			length=points1.length
			while (points1.length>0) && (index<length)
				index+=1
				pt=[]
				pt<<points1.shift
				(0..points1.length-1).each{|i|
					if ((pt[-1][1].distance(points1[i][0]).to_f) <= 0.00001)
						pt<< points1[i]
					elsif ((pt[0][0].distance(points1[i][1]).to_f) <= 0.00001)
						pt.unshift points1[i]
					else
						break
					end
				}
				points1=points1-pt
				points2<<pt
			end
			if points2.length>1 && (points2[-1][-1][1].distance(points2[0][0][0]).to_f <= 0.00001 )
				points2[0] =points2[-1]+points2[0]
				points2.delete_at(-1)
			end
			#p "asdf",points2.length
			pend_hash={}
			if points2.length==1
				if close
					if points2[0][0][0].distance(points2[0][-1][1]).to_f >= 0
						line1=[points2[0][0][0],(points2[0][0][1]-points2[0][0][0])]
						line2=[points2[0][-1][0],(points2[0][-1][1]-points2[0][-1][0])]
						if line1[1].length>0 && line2.length>0
							p= Geom::intersect_line_line(line1,line2)
							if p &&
								(((p.distance(points2[0][0][0]).to_f <=p.distance(points2[0][0][1]).to_f ) &&
								(points2[0][0][0].distance(points2[0][0][1]).to_f <= p.distance(points2[0][0][1]).to_f )))
									points2[0][0][0]=p
							end
							if p &&
								((p.distance(points2[0][-1][1]).to_f <= p.distance(points2[0][-1][0]).to_f ) &&
								(points2[0][-1][1].distance(points2[0][-1][0]).to_f <= p.distance(points2[0][-1][0]).to_f))
									points2[0][-1][1]=p
							end
						end
					end
				end
				ps1=points2[0]
				pt=[]
				return ps1 if !close
				#return ps1 if (ps1[-1][1].distance(ps1[0][0]).to_f <= 0.00001 )
				if ps1.length>0
					(0..ps1.length-1).each{|i1|
						line1=[ps1[i1][0],(ps1[i1][1]-ps1[i1][0])]
						pt1=[]
						ps2=points2[0]
						(0..ps2.length-1).each{|i2|
							if i1!=i2
								line2=[ps2[i2][0],(ps2[i2][1]-ps2[i2][0])]
								if line1[1].length!=0 && line2[1].length!=0
									p= Geom::intersect_line_line line1,line2
									if p  &&
										(p.distance(ps1[i1][0]) <= ps1[i1][0].distance(ps1[i1][1])) &&
										(p.distance(ps1[i1][1]) <= ps1[i1][0].distance(ps1[i1][1])) &&
										(p.distance(ps2[i2][0]) <= ps2[i2][0].distance(ps2[i2][1])) &&
										(p.distance(ps2[i2][1]) <= ps2[i2][0].distance(ps2[i2][1]))
											pt1<<[0,i1,p,0,i2]
									end
								end
							end
						}
						pt1=pt1.sort_by{|t| t[2].distance(ps1[t[1]][0])}
						if pt1.length>0
							pt+=pt1
						end
					}
					if pt.length>0
						pend_hash[0]=[] if !pend_hash[0]
						pend_hash[0]+=pt
					end
				end
			else
				#p points2.length
				(0..points2.length-1).each{|j|
					ps1=points2[j]
					no_continous=true
					pt=[]
					if ps1.length>0
						(0..ps1.length-1).each{|i1|
							line1=[ps1[i1][0],(ps1[i1][1]-ps1[i1][0])]
							pt1=[]
							(0..points2.length-1).each{|i|
								if i!=j
									ps2=(points2)[i]
									(0..ps2.length-1).each{|i2|
										line2=[ps2[i2][0],(ps2[i2][1]-ps2[i2][0])]
										if line1[1].length!=0 && line2[1].length!=0
											p= Geom::intersect_line_line line1,line2
											if p &&
												(p.distance(ps1[i1][0])<=ps1[i1][0].distance(ps1[i1][1])) &&
												(p.distance(ps1[i1][1])<=ps1[i1][0].distance(ps1[i1][1])) &&
												(p.distance(ps2[i2][0])<=ps2[i2][0].distance(ps2[i2][1])) &&
												(p.distance(ps2[i2][1])<=ps2[i2][0].distance(ps2[i2][1]))# &&
												 #p.distance(ps1[i1][1])>0 && p.distance(ps2[i2][0])>0 &&
												!point_in_points(points2,[j,i],p)
												pt1<<[j,i1,p,i,i2]
											end
										end
									}
								end
							}
							pt1=pt1.sort_by{|t| t[2].distance(ps1[t[1]][0])}
							if pt1.length>0
								pt+=pt1.clone
							end
						}
						if pt.length>0
							pend_hash[j]=[] if !pend_hash[j]
							pend_hash[j]+=pt.clone
						end
						# if !pend_hash[j] || pend_hash[j].empty?
							# return []
						# end
					end
				}
			end
			ps=[]
			pend=[]
			pend_hash.keys.each{|k|
				pend << pend_hash[k]
			}
			#p "len",pend[0].length,points2[0].length
			if close
				pend.each_index{|i|
					pp=pend[i]
					if pp.length%2==0
						check=true
						(0..pp.length-2).each{|j|
							if (j%2)==0
								pt=pp[j]
								pt1=pp[j+1]
								if (pt[0]==pt1[3]) && (pt[1]==pt1[4]) && (pt[3]==pt1[0]) && (pt[4]==pt1[1]) &&
									(pt[2].distance(pt1[2]).to_f <0.00001)
								else
									check=false
									break
								end
							end
							break if !check
						}
						if check
							ps=points2
						else
							(0..pp.length-2).each{|j|
								if (j%2)==0
									pt=pp[j]
									pt1=pp[j+1]
									ps1=points2[pt[0]][pt[1]..pt1[1]].collect{|p| p.clone}
									if ps1 && ps1.length>0
										ps1[0][0]=pt[2]
										ps1[-1][1]=pt1[2]
										ps<<ps1
									end
								end
							}
						end
					end
				}

			else
				#p "start" ,pend.length
				points2.each_index{|n|
					stop=false
					pend.each_index{|i|
						if pend[i][0][0]==n
							pp=pend[i]
							if n==0 && i==n
								pa=[i,0,points2[0][0][0]]
								pp.unshift pa
							elsif (n==(points2.length-1)) && i==n
								pa=[i,-1,points2[-1][-1][1]]
								pp<< pa
							end
							#p "ff",pp.length,"f",pp
							if pp.length%2==0
								(0..pp.length-2).each{|j|
									if (j%2)==0
										pt=pp[j]
										pt1=pp[j+1]
										if pt[0]==pt1[0]
											ps1=points2[pt[0]][pt[1]..pt1[1]].collect{|p| p.clone}
											if ps1 && ps1.length>0
												ps1[0][0]=pt[2]
												ps1[-1][1]=pt1[2]
												ps<<ps1
											end
										end
									end
								}
							# else
								# return []
							end
							stop=true
							break
						end
					}
					if !stop && points2[n].length>1
						ps<< points2[n]
					end
				}
			end
			return ps.flatten(1)
			rescue
			p "error edit"
			return points1
		end
		def self.point_in_points(points,ps,p)
			(0..points.length-1).each{|i|
				if !ps.include?(i) || ps.empty?
					pl=points[i]
					pl1=[]
					pl2=[]
					points[i].each{|pp|
						pl1+=pp[0..1]
						pl1+=[pp[2].position,pp[3].position]
					}
					polygon=pl1+pl2.reverse
					polygon.uniq!
					if Geom::point_in_polygon_2D(p,polygon,false)
						return true
					end
				end
			}
			return false
		end

		def self.find_weld(edges,check_hide_edge=true)
			ver_list_connected=[]
			if edges.length>0
				edges=edges.select{|e| !(e.hidden? || e.soft?)}.uniq if check_hide_edge
				edge_list_connected= find_all_connected edges
				ver_list_connected=edge_list_connected.collect{|eds| find_vertices(eds)}
			end
			return ver_list_connected
		end
		def self.find_all_connected(edges)
			edges1=edges.clone
			i=0;list=[]
			while (edges1.length>0) && (i<edges.length)
				ls=[]
				if f=edges1.find{|e| e.vertices[0].edges.length==1 || e.vertices[1].edges.length==1}
					ls<<f
					edges1.delete f
				else
					ls<<edges1.shift
				end
				allconnect= (ls.last.all_connected) & edges
				ls+=allconnect
				ls.uniq!
				list<< ls
				edges1-=ls
				i+=1
			end
			return list
			rescue =>error
			p error
			p error.backtrace.join('\n')
			return edges
		end
		def self.find_vertices(edges_new1)
			edges_new=edges_new1.clone
			i=0
			list=[]
			list1=[]
			while (edges_new.length>0) && (i<edges_new1.length)
				vers=edges_new.flat_map{|e| e.vertices}.uniq
				ls=[]
				if f=vers.find{|v| (l=(v.edges & edges_new1).length )&& l && ((l==1) ||(l>2) )}
					ls<<f
					vers.delete f
				elsif f=vers.find{|v| (l=(v.edges ).length )&& l && ((l==1) || (l>2))}
					ls<<f
					vers.delete f
				elsif f=vers.find{|v| (v.edges & edges_new).length==1}
					ls<<f
					vers.delete f
				else
					ls<< vers.shift
				end
				length=vers.length
				j=0
				while (edges_new.length>0) && (j<length) && (vers.length>0)
					j+=1
					vs=ls.last.edges.flat_map{|e| e.vertices} & vers
					lc=vs.find{|l|
							l.common_edge(ls.last) &&
							((l.edges & ls.last.edges) & edges_new).length==1
						}
					if lc
						led=(lc.edges & ls.last.edges) & edges_new
						edges_new-=led
						ls<< lc
						vers.delete(lc)
						break if ((lc.edges & edges_new1).length>2)
					end
				end
				if ls.length<=2
					list1 << ls
				else
					list << ls
				end
				i+=(ls.length-1)
			end

			j=0
			length=list1.length
			while list1.length>0 && j<length
					ls1=list1[0]
					list.each_index{|i|
						if ((ls1 & list[i]).length==2) && (((list[i][0].edges & list[i][-1].edges) & edges_new1).length==1)
							if (list[i][0].edges & edges_new1).length>(list[i][-1].edges & edges_new1).length
								list[i]<<list[i][0]
								list1.delete_at(0)
								break
							elsif (list[i][0].edges & edges_new1).length<(list[i][-1].edges & edges_new1).length
								list[i].unshift list[i][-1]
								list1.delete_at(0)
								break
							elsif (list[i][0].edges).length>(list[i][-1].edges).length
								list[i]<<list[i][0]
								list1.delete_at(0)
								break
							elsif (list[i][0].edges ).length<(list[i][-1].edges ).length
								list[i].unshift list[i][-1]
								list1.delete_at(0)
								break
							elsif ls1[0]==list[i][0] && ls1[-1]==list[i][-1] #&& ls1[0].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2 &&
								#ls1[1].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2
								list[i]<<ls1[0]
								list1.delete_at(0)
								break
							elsif ls1[0]==list[i][-1] && ls1[-1]==list[i][0] #&& ls1[0].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2 &&
								#ls1[1].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2
								list[i]<<ls1[-1]
								list1.delete_at(0)
								break
							end
						end
					}
				j+=1
			end
			return (list + list1)
		end
		def self.get_setting_model(name,length)
			model = Sketchup.active_model
			attrdicts = model.attribute_dictionaries
			suforyou=attrdicts['suforyou']
			if suforyou
				if suforyou[name] && (suforyou[name].length==length)
					@setting=suforyou[name].to_a
				end
			end
			@setting=Array.new(length,0)
			rescue
			@setting=Array.new(length,0)
		end
		def self.set_setting_model(name,value)
			model = Sketchup.active_model
			attrdicts = model.attribute_dictionaries
			model.set_attribute 'suforyou', name, value.to_s
		end
		def self.main(mode)
			if !@setting
				#@setting=get_setting_model(NAME,4)
				@setting=[0,0,0,0]
			end
			Sketchup.active_model.select_tool Offset.new()
			#set_setting_model(NAME,@setting)
		end
	end
end